@@ -1,13 +1,15 @@
-"""This module parses and generates contentlines as defined in RFC 2445
+"""This module parses and generates contentlines as defined in RFC 5545
 (iCalendar), but will probably work for other MIME types with similar syntax.
 Eg. RFC 2426 (vCard)
 
 It is stupid in the sense that it treats the content purely as strings. No type
 conversion is attempted.
 """
+from __future__ import annotations
 
+import os
 from icalendar.caselessdict import CaselessDict
-from icalendar.parser_tools import DEFAULT_ENCODING
+from icalendar.parser_tools import DEFAULT_ENCODING, ICAL_TYPE
 from icalendar.parser_tools import SEQUENCE_TYPES
 from icalendar.parser_tools import to_unicode
 
@@ -15,51 +17,43 @@ import re
 
 
 def escape_char(text):
-    """Format value according to iCalendar TEXT escaping rules.
-    """
+    """Format value according to iCalendar TEXT escaping rules."""
     assert isinstance(text, (str, bytes))
     # NOTE: ORDER MATTERS!
-    return text.replace(r'\N', '\n')\
-               .replace('\\', '\\\\')\
-               .replace(';', r'\;')\
-               .replace(',', r'\,')\
-               .replace('\r\n', r'\n')\
-               .replace('\n', r'\n')
+    return (
+        text.replace(r"\N", "\n")
+        .replace("\\", "\\\\")
+        .replace(";", r"\;")
+        .replace(",", r"\,")
+        .replace("\r\n", r"\n")
+        .replace("\n", r"\n")
+    )
 
 
 def unescape_char(text):
     assert isinstance(text, (str, bytes))
     # NOTE: ORDER MATTERS!
     if isinstance(text, str):
-        return text.replace('\\N', '\\n')\
-                   .replace('\r\n', '\n')\
-                   .replace('\\n', '\n')\
-                   .replace('\\,', ',')\
-                   .replace('\\;', ';')\
-                   .replace('\\\\', '\\')
+        return (
+            text.replace("\\N", "\\n")
+            .replace("\r\n", "\n")
+            .replace("\\n", "\n")
+            .replace("\\,", ",")
+            .replace("\\;", ";")
+            .replace("\\\\", "\\")
+        )
     elif isinstance(text, bytes):
-        return text.replace(b'\\N', b'\\n')\
-                   .replace(b'\r\n', b'\n')\
-                   .replace(b'\n', b'\n')\
-                   .replace(b'\\,', b',')\
-                   .replace(b'\\;', b';')\
-                   .replace(b'\\\\', b'\\')
+        return (
+            text.replace(b"\\N", b"\\n")
+            .replace(b"\r\n", b"\n")
+            .replace(b"\\n", b"\n")
+            .replace(b"\\,", b",")
+            .replace(b"\\;", b";")
+            .replace(b"\\\\", b"\\")
+        )
 
 
-def tzid_from_dt(dt):
-    tzid = None
-    if hasattr(dt.tzinfo, 'zone'):
-        tzid = dt.tzinfo.zone  # pytz implementation
-    elif hasattr(dt.tzinfo, 'key'):
-        tzid = dt.tzinfo.key  # ZoneInfo implementation
-    elif hasattr(dt.tzinfo, 'tzname'):
-        # dateutil implementation, but this is broken
-        # See https://github.com/collective/icalendar/issues/333 for details
-        tzid = dt.tzinfo.tzname(dt)
-    return tzid
-
-
-def foldline(line, limit=75, fold_sep='\r\n '):
+def foldline(line, limit=75, fold_sep="\r\n "):
     """Make a string folded as defined in RFC5545
     Lines of text SHOULD NOT be longer than 75 octets, excluding the line
     break.  Long content lines SHOULD be split into a multiple line
@@ -69,16 +63,16 @@ def foldline(line, limit=75, fold_sep='\r\n '):
     SPACE or HTAB).
     """
     assert isinstance(line, str)
-    assert '\n' not in line
+    assert "\n" not in line
 
     # Use a fast and simple variant for the common case that line is all ASCII.
     try:
-        line.encode('ascii')
+        line.encode("ascii")
     except (UnicodeEncodeError, UnicodeDecodeError):
         pass
     else:
         return fold_sep.join(
-            line[i:i + limit - 1] for i in range(0, len(line), limit - 1)
+            line[i : i + limit - 1] for i in range(0, len(line), limit - 1)
         )
 
     ret_chars = []
@@ -91,34 +85,33 @@ def foldline(line, limit=75, fold_sep='\r\n '):
             byte_count = char_byte_len
         ret_chars.append(char)
 
-    return ''.join(ret_chars)
+    return "".join(ret_chars)
 
 
 #################################################################
 # Property parameter stuff
 
-def param_value(value):
-    """Returns a parameter value.
-    """
+
+def param_value(value, always_quote=False):
+    """Returns a parameter value."""
     if isinstance(value, SEQUENCE_TYPES):
-        return q_join(value)
-    elif isinstance(value, str):
-        return dquote(value)
-    else:
-        return dquote(value.to_ical().decode(DEFAULT_ENCODING))
+        return q_join(map(rfc_6868_escape, value), always_quote=always_quote)
+    if isinstance(value, str):
+        return dquote(rfc_6868_escape(value), always_quote=always_quote)
+    return dquote(rfc_6868_escape(value.to_ical().decode(DEFAULT_ENCODING)))
 
 
 # Could be improved
 
 # [\w-] because of the iCalendar RFC
 # . because of the vCard RFC
-NAME = re.compile(r'[\w.-]+')
+NAME = re.compile(r"[\w.-]+")
 
-UNSAFE_CHAR = re.compile('[\x00-\x08\x0a-\x1f\x7F",:;]')
-QUNSAFE_CHAR = re.compile('[\x00-\x08\x0a-\x1f\x7F"]')
-FOLD = re.compile(b'(\r?\n)+[ \t]')
-uFOLD = re.compile('(\r?\n)+[ \t]')
-NEWLINE = re.compile(r'\r?\n')
+UNSAFE_CHAR = re.compile('[\x00-\x08\x0a-\x1f\x7f",:;]')
+QUNSAFE_CHAR = re.compile('[\x00-\x08\x0a-\x1f\x7f"]')
+FOLD = re.compile(b"(\r?\n)+[ \t]")
+uFOLD = re.compile("(\r?\n)+[ \t]")
+NEWLINE = re.compile(r"\r?\n")
 
 
 def validate_token(name):
@@ -139,21 +132,19 @@ def validate_param_value(value, quoted=True):
 QUOTABLE = re.compile("[,;: ’']")
 
 
-def dquote(val):
-    """Enclose parameter values containing [,;:] in double quotes.
-    """
+def dquote(val, always_quote=False):
+    """Enclose parameter values containing [,;:] in double quotes."""
     # a double-quote character is forbidden to appear in a parameter value
     # so replace it with a single-quote character
     val = val.replace('"', "'")
-    if QUOTABLE.search(val):
+    if QUOTABLE.search(val) or always_quote:
         return f'"{val}"'
     return val
 
 
 # parsing helper
-def q_split(st, sep=',', maxsplit=-1):
-    """Splits a string on char, taking double (q)uotes into considderation.
-    """
+def q_split(st, sep=",", maxsplit=-1):
+    """Splits a string on char, taking double (q)uotes into considderation."""
     if maxsplit == 0:
         return [st]
 
@@ -175,41 +166,50 @@ def q_split(st, sep=',', maxsplit=-1):
     return result
 
 
-def q_join(lst, sep=','):
-    """Joins a list on sep, quoting strings with QUOTABLE chars.
-    """
-    return sep.join(dquote(itm) for itm in lst)
+def q_join(lst, sep=",", always_quote=False):
+    """Joins a list on sep, quoting strings with QUOTABLE chars."""
+    return sep.join(dquote(itm, always_quote=always_quote) for itm in lst)
 
 
 class Parameters(CaselessDict):
     """Parser and generator of Property parameter strings. It knows nothing of
     datatypes. Its main concern is textual structure.
     """
+    
+    # The following paremeters must always be enclosed in double quotes
+    always_quoted = (
+        "ALTREP",
+        "DELEGATED-FROM",
+        "DELEGATED-TO",
+        "DIR",
+        "MEMBER",
+        "SENT-BY",
+    )
 
     def params(self):
-        """In rfc2445 keys are called parameters, so this is to be consitent
+        """In RFC 5545 keys are called parameters, so this is to be consitent
         with the naming conventions.
         """
         return self.keys()
 
-# TODO?
-# Later, when I get more time... need to finish this off now. The last major
-# thing missing.
-#   def _encode(self, name, value, cond=1):
-#       # internal, for conditional convertion of values.
-#       if cond:
-#           klass = types_factory.for_property(name)
-#           return klass(value)
-#       return value
-#
-#   def add(self, name, value, encode=0):
-#       "Add a parameter value and optionally encode it."
-#       if encode:
-#           value = self._encode(name, value, encode)
-#       self[name] = value
-#
-#   def decoded(self, name):
-#       "returns a decoded value, or list of same"
+    # TODO?
+    # Later, when I get more time... need to finish this off now. The last major
+    # thing missing.
+    #   def _encode(self, name, value, cond=1):
+    #       # internal, for conditional convertion of values.
+    #       if cond:
+    #           klass = types_factory.for_property(name)
+    #           return klass(value)
+    #       return value
+    #
+    #   def add(self, name, value, encode=0):
+    #       "Add a parameter value and optionally encode it."
+    #       if encode:
+    #           value = self._encode(name, value, encode)
+    #       self[name] = value
+    #
+    #   def decoded(self, name):
+    #       "returns a decoded value, or list of same"
 
     def to_ical(self, sorted=True):
         result = []
@@ -218,13 +218,13 @@ class Parameters(CaselessDict):
             items.sort()
 
         for key, value in items:
-            value = param_value(value)
-            if isinstance(value, str):
-                value = value.encode(DEFAULT_ENCODING)
+            upper_key = key.upper()
+            quoted_value = param_value(value, always_quote=upper_key in self.always_quoted)
+            if isinstance(quoted_value, str):
+                quoted_value = quoted_value.encode(DEFAULT_ENCODING)
             # CaselessDict keys are always unicode
-            key = key.upper().encode(DEFAULT_ENCODING)
-            result.append(key + b'=' + value)
-        return b';'.join(result)
+            result.append(upper_key.encode(DEFAULT_ENCODING) + b"=" + quoted_value)
+        return b";".join(result)
 
     @classmethod
     def from_ical(cls, st, strict=False):
@@ -232,24 +232,24 @@ class Parameters(CaselessDict):
 
         # parse into strings
         result = cls()
-        for param in q_split(st, ';'):
+        for param in q_split(st, ";"):
             try:
-                key, val = q_split(param, '=', maxsplit=1)
+                key, val = q_split(param, "=", maxsplit=1)
                 validate_token(key)
                 # Property parameter values that are not in quoted
                 # strings are case insensitive.
                 vals = []
-                for v in q_split(val, ','):
+                for v in q_split(val, ","):
                     if v.startswith('"') and v.endswith('"'):
                         v = v.strip('"')
                         validate_param_value(v, quoted=True)
-                        vals.append(v)
+                        vals.append(rfc_6868_unescape(v))
                     else:
                         validate_param_value(v, quoted=False)
                         if strict:
-                            vals.append(v.upper())
+                            vals.append(rfc_6868_unescape(v.upper()))
                         else:
-                            vals.append(v)
+                            vals.append(rfc_6868_unescape(v))
                 if not vals:
                     result[key] = val
                 else:
@@ -258,22 +258,66 @@ class Parameters(CaselessDict):
                     else:
                         result[key] = vals
             except ValueError as exc:
-                raise ValueError(
-                    f'{param!r} is not a valid parameter string: {exc}')
+                raise ValueError(f"{param!r} is not a valid parameter string: {exc}")
         return result
 
 
 def escape_string(val):
     # f'{i:02X}'
-    return val.replace(r'\,', '%2C').replace(r'\:', '%3A')\
-              .replace(r'\;', '%3B').replace(r'\\', '%5C')
+    return (
+        val.replace(r"\,", "%2C")
+        .replace(r"\:", "%3A")
+        .replace(r"\;", "%3B")
+        .replace(r"\\", "%5C")
+    )
 
 
 def unescape_string(val):
-    return val.replace('%2C', ',').replace('%3A', ':')\
-              .replace('%3B', ';').replace('%5C', '\\')
+    return (
+        val.replace("%2C", ",")
+        .replace("%3A", ":")
+        .replace("%3B", ";")
+        .replace("%5C", "\\")
+    )
 
 
+RFC_6868_UNESCAPE_REGEX = re.compile(r"\^\^|\^n|\^'")
+
+
+def rfc_6868_unescape(param_value: str) -> str:
+    """Take care of :rfc:`6868` unescaping.
+
+    - ^^ -> ^
+    - ^n -> system specific newline
+    - ^' -> "
+    - ^ with others stay intact
+    """
+    replacements = {
+        "^^": "^",
+        "^n": os.linesep,
+        "^'": '"',
+    }
+    return RFC_6868_UNESCAPE_REGEX.sub(lambda m: replacements.get(m.group(0), m.group(0)), param_value)
+
+
+RFC_6868_ESCAPE_REGEX = re.compile(r'\^|\r\n|\r|\n|"')
+
+def rfc_6868_escape(param_value: str) -> str:
+    """Take care of :rfc:`6868` escaping.
+
+    - ^ -> ^^
+    - " -> ^'
+    - newline -> ^n
+    """
+    replacements = {
+        "^": "^^",
+        "\n": "^n",
+        "\r": "^n",
+        "\r\n": "^n",
+        '"': "^'",
+    }
+    return RFC_6868_ESCAPE_REGEX.sub(lambda m: replacements.get(m.group(0), m.group(0)), param_value)
+
 def unescape_list_or_string(val):
     if isinstance(val, list):
         return [unescape_string(s) for s in val]
@@ -284,26 +328,29 @@ def unescape_list_or_string(val):
 #########################################
 # parsing and generation of content lines
 
+
 class Contentline(str):
     """A content line is basically a string that can be folded and parsed into
     parts.
     """
+
     def __new__(cls, value, strict=False, encoding=DEFAULT_ENCODING):
         value = to_unicode(value, encoding=encoding)
-        assert '\n' not in value, ('Content line can not contain unescaped '
-                                   'new line characters.')
+        assert "\n" not in value, (
+            "Content line can not contain unescaped " "new line characters."
+        )
         self = super().__new__(cls, value)
         self.strict = strict
         return self
 
     @classmethod
-    def from_parts(cls, name, params, values, sorted=True):
-        """Turn a parts into a content line.
-        """
+    def from_parts(cls, name:ICAL_TYPE, params: Parameters, values, sorted=True):
+        """Turn a parts into a content line."""
         assert isinstance(params, Parameters)
-        if hasattr(values, 'to_ical'):
+        if hasattr(values, "to_ical"):
             values = values.to_ical()
         else:
+            from icalendar.prop import vText
             values = vText(values).to_ical()
         # elif isinstance(values, basestring):
         #    values = escape_char(values)
@@ -314,12 +361,11 @@ class Contentline(str):
         values = to_unicode(values)
         if params:
             params = to_unicode(params.to_ical(sorted=sorted))
-            return cls(f'{name};{params}:{values}')
-        return cls(f'{name}:{values}')
+            return cls(f"{name};{params}:{values}")
+        return cls(f"{name}:{values}")
 
     def parts(self):
-        """Split the content line up into (name, parameters, values) parts.
-        """
+        """Split the content line up into (name, parameters, values) parts."""
         try:
             st = escape_string(self)
             name_split = None
@@ -327,39 +373,40 @@ class Contentline(str):
             in_quotes = False
             for i, ch in enumerate(st):
                 if not in_quotes:
-                    if ch in ':;' and not name_split:
+                    if ch in ":;" and not name_split:
                         name_split = i
-                    if ch == ':' and not value_split:
+                    if ch == ":" and not value_split:
                         value_split = i
                 if ch == '"':
                     in_quotes = not in_quotes
             name = unescape_string(st[:name_split])
             if not name:
-                raise ValueError('Key name is required')
+                raise ValueError("Key name is required")
             validate_token(name)
             if not value_split:
                 value_split = i + 1
             if not name_split or name_split + 1 == value_split:
-                raise ValueError('Invalid content line')
-            params = Parameters.from_ical(st[name_split + 1: value_split],
-                                          strict=self.strict)
+                raise ValueError("Invalid content line")
+            params = Parameters.from_ical(
+                st[name_split + 1 : value_split], strict=self.strict
+            )
             params = Parameters(
                 (unescape_string(key), unescape_list_or_string(value))
                 for key, value in iter(params.items())
             )
-            values = unescape_string(st[value_split + 1:])
+            values = unescape_string(st[value_split + 1 :])
             return (name, params, values)
         except ValueError as exc:
             raise ValueError(
-                f"Content line could not be parsed into parts: '{self}': {exc}")
+                f"Content line could not be parsed into parts: '{self}': {exc}"
+            ) from exc
 
     @classmethod
     def from_ical(cls, ical, strict=False):
-        """Unfold the content lines in an iCalendar into long content lines.
-        """
+        """Unfold the content lines in an iCalendar into long content lines."""
         ical = to_unicode(ical)
         # a fold is carriage return followed by either a space or a tab
-        return cls(uFOLD.sub('', ical), strict=strict)
+        return cls(uFOLD.sub("", ical), strict=strict)
 
     def to_ical(self):
         """Long content lines are folded so they are less than 75 characters
@@ -375,25 +422,46 @@ class Contentlines(list):
     """
 
     def to_ical(self):
-        """Simply join self.
-        """
-        return b'\r\n'.join(line.to_ical() for line in self if line) + b'\r\n'
+        """Simply join self."""
+        return b"\r\n".join(line.to_ical() for line in self if line) + b"\r\n"
 
     @classmethod
     def from_ical(cls, st):
-        """Parses a string into content lines.
-        """
+        """Parses a string into content lines."""
         st = to_unicode(st)
         try:
             # a fold is carriage return followed by either a space or a tab
-            unfolded = uFOLD.sub('', st)
-            lines = cls(Contentline(line) for
-                        line in NEWLINE.split(unfolded) if line)
-            lines.append('')  # '\r\n' at the end of every content line
+            unfolded = uFOLD.sub("", st)
+            lines = cls(Contentline(line) for line in NEWLINE.split(unfolded) if line)
+            lines.append("")  # '\r\n' at the end of every content line
             return lines
         except Exception:
-            raise ValueError('Expected StringType with content lines')
+            raise ValueError("Expected StringType with content lines")
 
 
-# XXX: what kind of hack is this? import depends to be at end
-from icalendar.prop import vText
+__all__ = [
+    "Contentline",
+    "Contentlines",
+    "FOLD",
+    "NAME",
+    "NEWLINE",
+    "Parameters",
+    "QUNSAFE_CHAR",
+    "QUOTABLE",
+    "UNSAFE_CHAR",
+    "dquote",
+    "escape_char",
+    "escape_string",
+    "foldline",
+    "param_value",
+    "q_join",
+    "q_split",
+    "rfc_6868_escape",
+    "rfc_6868_unescape",
+    "uFOLD",
+    "unescape_char",
+    "unescape_list_or_string",
+    "unescape_string",
+    "validate_param_value",
+    "validate_token",
+]
